

ROCK = 0
ROOM = 1

-- oneliner stuff
function isOfKind(x,y,v) return grid[x][y] == v end
function isInside(x,y) return (x>=1 and x<=xSize and y>=1 and y<=ySize) end
function randCoord() return math.random(1,xSize), math.random(1,ySize) end

-- check, set, count
function setRoom(x, y, v)
	if v == ROCK then -- fill with rock
		if grid[x][y] ~= ROCK then
			grid[x][y] = v
			usedCell = usedCell - 1
		else print("Fail; trying to fill a solid cell!")
		end
	elseif v == ROOM then -- dig out a room
		if grid[x][y] ~= ROOM then
			grid[x][y] = v
			usedCell = usedCell + 1
		else print("Fail; trying to dig an empty cell!")
		end
	elseif v == 2 then -- for floodfill
		if grid[x][y] == ROOM then
			grid[x][y] = v
		else print("Fail; trying to check "..grid[x][y].." instead of 1 (ROOM)!")
		end
	end
end

-- find a cell of type not-v, but has at least one neighbour with v
function findNewStart(v)
	local x, y, check = 1, 1, false
	repeat
		x, y = randCoord()
		check = (not isOfKind(x, y, v)) and (findAround(x, y, v) > 0)
	until check
	return x, y
end

-- converts a dir-String into x/y offsets
function getDir(str)
	local rx, ry = 0, 0
	if 			str == 'U' then rx, ry = 0	, -1
	elseif 	str == 'D' then rx, ry = 0	, 1
	elseif 	str == 'L' then rx, ry = -1	, 0
	elseif 	str == 'R' then rx, ry = 1 	, 0
	else rx, ry = nil, nil
	end
	return rx, ry
end

-- gives random direction in x/y offsets
function newRandDir()
	local d = {'U', 'D', 'L', 'R'}
	return getDir(d[math.random(1, 4)])
end

-- checks neighbours, returns number of desired target and table of directions where to find these
function findAround(x, y, v)
	local res = {}
	if y > 1 					and grid[x][y - 1] == v then table.insert(res, 'U') end
	if y < ySize - 1	and grid[x][y + 1] == v then table.insert(res, 'D') end
	if x > 1 					and grid[x - 1][y] == v then table.insert(res, 'L') end
	if x < xSize - 1	and grid[x + 1][y] == v then table.insert(res, 'R') end
	return #res, res
end

-- checks diagonal pseudo-neighbours, returns number of finds
function findDiag(x, y, v)
	local res = 0
	if x > 1 					and y > 1 					and isOfKind(x - 1, y - 1, v) then res = res + 1 end
	if x < xSize - 1 	and y > 1 					and isOfKind(x + 1, y - 1, v) then res = res + 1 end
	if x > 1 					and y < ySize - 1 	and isOfKind(x - 1, y + 1, v) then res = res + 1 end
	if x < xSize - 1 	and y < ySize - 1 	and isOfKind(x + 1, y + 1, v) then res = res + 1 end
	return res
end

-- checks if the given area is free of rocks
function isVoidArea(sx, sy, ex, ey)
	local isFree = true
	for _x = sx, ex do for _y = sy, ey do
		if isOfKind(_x, _y, ROCK) then isFree = false end
	end end
	return isFree
end

-- fills out irregular-shaped empty area, inside a void rectangle
function carveArea(sx, sy, ex, ey)

	local xL, yL = ex - sx + 1, ey - sy + 1
	for _x = sx, ex do for _y = sy, ey do
		setRoom(_x, _y, ROCK)
	end end
	if math.random(2) == 1 then setRoom(sx, sy, ROOM)
		if math.random(2) == 1 then setRoom(sx + 1, sy, ROOM) end
		if math.random(2) == 1 then setRoom(sx, sy + 1, ROOM) end
	end
	if math.random(2) == 1 then setRoom(sx, ey, ROOM)
		if math.random(2) == 1 then setRoom(sx + 1, ey, ROOM) end
		if math.random(2) == 1 then setRoom(sx, ey - 1, ROOM) end
	end
	if math.random(2) == 1 then setRoom(ex, sy, ROOM)
		if math.random(2) == 1 then setRoom(ex - 1, sy, ROOM) end
		if math.random(2) == 1 then setRoom(ex, sy + 1, ROOM) end
	end
	if math.random(2) == 1 then setRoom(ex, ey, ROOM)
		if math.random(2) == 1 then setRoom(ex - 1, ey, ROOM) end
		if math.random(2) == 1 then setRoom(ex, ey - 1, ROOM) end
	end
end





------------
-- dig rooms
------------

function digIt()

	if ratio < maxRatio then
		local wx, wy = sx, sy -- wx/wy (w:walk)

		local nbCount, nbTable = findAround(wx, wy, ROCK)
		while nbCount > 0 do
			setRoom(wx, wy, ROOM)
			newDir = nbTable[math.random(1, nbCount)]
			local dx, dy = getDir(newDir)
			wx = wx + dx; wy = wy + dy
			nbCount, nbTable = findAround(wx, wy, ROCK)
		end

		local success = false
		local dx, dy = newRandDir()
		repeat -- finds a "wall"
			local dx, dy = newRandDir()
			wx = wx + dx; wy = wy + dy
			if not isInside(wx, wy) then wx, wy = findNewStart(ROOM) -- find rock, with room as neighbour
			 else if isOfKind(wx, wy, ROCK) then success = true end
			end
		until success

		sx, sy = wx, wy -- wall is found, then set new seed and back to beginning
	end
end


--------------------------------------------------
-- erases isolated 1x1 rock islands inside rooms
--------------------------------------------------

function fillMiniholes()
	for _x = 1, xSize do for _y = 1, ySize do
		if isOfKind(_x, _y, ROCK) and findAround(_x, _y, ROOM) == 4 then setRoom(_x, _y, ROOM) end
	end end
end


------------------------------
-- some basic refinements
-- after growing phase is done
------------------------------

function refineIt()


	local sx, sy = randCoord()
	local ex, ey = sx + 2, sy + 2
	if isInside(ex, ey) and isVoidArea(sx, sy, ex, ey) then

		failcounter = 0
		local fail = {}; for _ = 1, 4 do fail[_] = false end
			for expandDir = 1, 4 do

				repeat
					if expandDir == 1 then
						if sx > 1 then
							sx = sx - 1
							if not isVoidArea(sx, sy, ex, ey) then sx = sx + 1; fail[1] = true end
						else fail[1] = true end
					end
					if expandDir == 2 then
						if sy > 1 then
							sy = sy - 1
							if not isVoidArea(sx, sy, ex, ey) then sy = sy + 1; fail[2] = true end
						else fail[2] = true end
					end
					if expandDir == 3 then
						if ex < xSize then
							ex = ex + 1
							if not isVoidArea(sx, sy, ex, ey) then ex = ex - 1; fail[3] = true end
						else fail[3] = true end
					end
					if expandDir == 4 then
						if ey < ySize then
							ey = ey + 1
							if not isVoidArea(sx, sy, ex, ey) then ey = ey - 1; fail[4] = true end
						else fail[4] = true end
					end
				until fail[expandDir]

			end -- for: 4x expandDir

		carveArea(sx + 1, sy + 1, ex - 1, ey - 1)
		

	end

end


------------------------
-- floodfill the rooms
-- to check connectivity
------------------------

function fillIt()
	
	local sx, sy, stack = 0, 0, {}
	local old, new = ROOM, 2 -- 2 is the filled room value (temporary)

	repeat -- find a room tile to start with
		sx, sy = randCoord()
	until isOfKind(sx, sy, ROOM)
	
	stack[#stack+1] = {sx, sy} -- push
	while #stack > 0 do
		local last = #stack
		sx = stack[last][1]; sy = stack[last][2]; stack[last] = nil -- pop
		if isOfKind(sx, sy, old) then
			setRoom(sx, sy, new)
			stack[#stack+1] = {sx + 1, sy}
			stack[#stack+1] = {sx - 1, sy}
			stack[#stack+1] = {sx, sy + 1}
			stack[#stack+1] = {sx, sy - 1}
		end
	end

	checkIt()

end


--------------------------------------
-- control if all is filled,
-- fix if not, and repeat fill & check
--------------------------------------

function checkIt()

	state = 'all connected' -- at start in good hope

	for _x = 1, xSize do for _y = 1, ySize do
		if isOfKind(_x, _y, ROOM) then
			state = 'disconnected parts'
			if findDiag(_x, _y, 2) > 0 then
				for ix = _x - 1, _x + 1 do for iy = _y - 1, _y + 1 do -- the dirty quick-fix loop
					if isInside(ix, iy) and isOfKind(ix, iy, ROCK) then grid[ix][iy] = 2 end
				end	end
			end
		end
	end end

	if state == 'disconnected parts' then
		for _x = 1, xSize do for _y = 1, ySize do -- clean all temporary rooms ("2"), so that fill works again
			if isOfKind(_x, _y, 2) then grid[_x][_y] = ROOM end -- override setroom, since we dont mess with cellcount here
		end end
		fillIt() -- had to fix, so fill again in case it wasnt the only problem
	end

end 






-------------
-- löve stuff
-------------

function love.load()
	io.stdout:setvbuf("no")

	state = 'grow' -- do something right after program start
	xSize = 100; ySize = 100

	cellSize = 8; cellOffset = math.ceil(cellSize/2) -- display related stuff

	maxCell = xSize * ySize -- cell counting
	usedCell = 0
	maxRatio = 0.15 -- stop filling here

	fStr = "LMB to restart | RMB to quit | [SPACE] to proceed\nCells used: %3d%% (stop at: %3d%%)"
	love.graphics.setMode(xSize * cellSize, ySize * cellSize)
	
	grid = {} -- "world" init
	for x = 0, xSize+1 do
		grid[x] = {}
		for y = 0, ySize+1 do
			grid[x][y] = 0
		end
	end

	sx, sy = math.ceil(xSize/2), math.ceil(ySize/2) -- lets start in middle
	newDir = {}
end



------------------------
function love.update(dt)
	ratio = usedCell / maxCell
	if state == 'grow' then digIt()
	elseif state == 'fill miniholes' then fillMiniholes()
	elseif state == 'refine' then refineIt()
	elseif state == 'floodfill' then fillIt()
	end
end
---


function love.draw()
	for x = 1, xSize do
		for y = 1, ySize do
			local sx = x*cellSize; local sy = y*cellSize
			local c = grid[x][y]
			if c == 0 then
				love.graphics.setPoint(cellSize+.5, 'smooth')
				love.graphics.setColor(40,50,0)
				love.graphics.point(sx + cellOffset, sy + cellOffset)
			else
				love.graphics.setPoint(cellSize*.8, 'rough')
				love.graphics.setColor(c*60,c*60,c*60)
				love.graphics.point(sx + cellOffset, sy + cellOffset)
			end
		end
	end	
	local stats = string.format(fStr..'\n... '..state, tostring(ratio*100), tostring(maxRatio*100))
	love.graphics.setColor(0,0,0)
	for _x = 3, 7 do for _y = 3, 7 do love.graphics.print(stats, _x, _y) end end
	love.graphics.setColor(230,20,20); love.graphics.print(stats, 5, 5)
end


function love.mousepressed(x, y, button)
	if button == 'r' then love.quit() end
	if button == 'l' then love.load() end
end


function love.keypressed(key, unicode)
	if key == ' ' then
		if 			state == 'grow' 					then state = 'fill miniholes'
		elseif 	state == 'fill miniholes' then state = 'refine'
		elseif 	state == 'refine' 				then state = 'floodfill'
		elseif 	state == 'floodfill' 			then state = 'done'
		elseif 	state == 'all connected' 	then love.load()
		end
	end
end


function love.keyreleased(key, unicode)
end

function love.mousereleased(x, y, button)
end

function love.focus(f)
end

function love.quit()
  print("Thanks for playing! Come back soon!")
  os.exit()
end



